import base64
import hashlib
from zlib import crc32

from fmtr.tools.config import ToolsConfig

SPECIALS = {'O': '9', '=': '9', 'I': '9'}

def hash_unit(value: str) -> float:
    """

    Hash the input string to a value between 0.0 and 1.0 (not secure).

    """
    value = str(value).encode(ToolsConfig.ENCODING)
    return float(crc32(value) & 0xffffffff) / 2 ** 32


def get_hash_readable(string, length=None):
    """

    Get hash optimised for information density and readability.

    """
    hash_sha1 = hashlib.sha1(string.encode()).digest()
    hash_b32 = base64.b32encode(hash_sha1).decode()
    value = hash_b32[:length]
    for old, new in SPECIALS.items():
        value = value.replace(old, new)
    return value



from itertools import chain
from typing import List, Dict, Any


def enlist(value) -> List[Any]:
    """

    Make a non-list into a singleton list

    """
    enlisted = value if isinstance(value, list) else [value]
    return enlisted


def dict_records_to_lists(data: List[Dict[Any, Any]], missing: Any = None) -> Dict[Any, List[Any]]:
    """

    Convert a list of dictionaries to lists format

    """
    keys = set(chain.from_iterable([datum.keys() for datum in data]))
    as_lists = {key: [] for key in keys}
    for datum in data:
        for key in keys:
            as_lists[key].append(datum.get(key, missing))
    return as_lists


def get_batch_sizes(total, num_batches):
    """

    Calculate the sizes of batches for a given total number of items and number of batches.

    """
    return [total // num_batches + (1 if x < total % num_batches else 0) for x in range(num_batches)]


def chunk_data(data, size: int):
    """

    Chunk data into batches of a given size, plus any remainder

    """
    chunked = [data[offset:offset + size] for offset in range(0, len(data), size)]
    return chunked



import datetime

from fmtr.tools.config_tools import ConfigClass


class ToolsConfig(ConfigClass):
    ENCODING = 'UTF-8'
    LIBRARY_NAME = 'fmtr.tools'
    DATE_FILENAME_FORMAT = '%Y-%m-%d'
    TIME_FILENAME_FORMAT = '%H-%M-%S'

    DATETIME_SEMVER_BUILD_FORMAT = f'{DATE_FILENAME_FORMAT}-{TIME_FILENAME_FORMAT}'
    DATETIME_FILENAME_FORMAT = f'{DATE_FILENAME_FORMAT}@{TIME_FILENAME_FORMAT}'
    DATETIME_NOW = datetime.datetime.now(datetime.timezone.utc)
    DATETIME_NOW_STR = DATETIME_NOW.strftime(DATETIME_FILENAME_FORMAT)
    SERIALIZATION_INDENT = 4

    FMTR_LOG_LEVEL_KEY = 'FMTR_LOG_LEVEL'



from time import sleep
from typing import ClassVar

import streamlit as st

from fmtr.tools.data_modelling_tools import Base
from fmtr.tools.logging_tools import logger
from fmtr.tools.path_tools import Path


class Interface(Base):
    """

    Base for using streamlit via classes

    """

    PATH: ClassVar = __file__
    LAYOUT: ClassVar = 'centered'

    parent: Base = None
    st: ClassVar = st

    def set_title(self):
        """

        Set page title and layout when root interface

        """

        self.st.set_page_config(page_title=self.NAME, layout=self.LAYOUT)
        self.st.title(self.NAME)

    def render(self):
        """

        Render the Interface

        """
        raise NotImplementedError()

    def get_key(self, seg=None):
        """

        Get a structure-friendly unique ID

        """
        if self.parent is None:
            base = Path()
        else:
            base = self.parent.get_key() / str(id(self))

        if seg:
            path = base / seg
        else:
            path = base

        return path

    def to_tabs(self, *classes):
        """

        Add tabs from a list of interface classes

        """
        tab_names = [cls.NAME for cls in classes]
        tabs = st.tabs(tab_names)

        for cls, tab in zip(classes, tabs):
            with tab:
                cls()

    @classmethod
    def is_streamlit(cls):
        """

        Infer whether we are running within StreamLit

        """
        return bool(st.context.headers)

    @classmethod
    @st.cache_resource(show_spinner=False)
    def get_state(cls):
        """

        Initialise this Interface and keep cached

        """
        msg = f'Initialising State: for App {cls.NAME}...'
        logger.info(msg)
        self = cls()
        return self

    @classmethod
    def launch(cls):
        """

        Launch StreamLit, if not already running - otherwise get self from cache and render

        """
        if cls.is_streamlit():
            self = cls.get_state()
            self.set_title()
            self.render()
        else:
            from streamlit.web import bootstrap
            bootstrap.run(cls.PATH, False, [], {})


class InterfaceTest(Interface):
    NAME: ClassVar = 'Test Interface'

    parent: Base = None

    def render(self):
        """

        Render the Interface

        """
        if not self.st.button('Run Test'):
            return
        msg = 'Running test...'
        with self.st.spinner(msg):
            sleep(3)
        self.st.success("Success!")


if __name__ == '__main__':
    InterfaceTest.launch()



import sys

from fmtr.tools.config import ToolsConfig
from fmtr.tools.config_tools import ConfigClass
from fmtr.tools.environment_tools import get
from fmtr.tools.path_tools import Path


class LoggingConfig(ConfigClass):
    SEP = ' '
    TIME = '<bold><green>{time:' + ToolsConfig.DATETIME_FILENAME_FORMAT + '}</green></bold>'
    ICON = '<level>{level.icon}</level>'
    LEVEL = '<level>{level:<8}</level>'
    FILE = '{file}:{line}'
    FUNCTION = '{function}(…)'
    MESSAGE = '{message}'
    DEFAULT_LEVEL_KEY = 'FMTR_LOG_LEVEL'
    DEFAULT_LEVEL = get(DEFAULT_LEVEL_KEY, 'INFO')

    FILENAME = f'log-{ToolsConfig.DATETIME_NOW_STR}.log'


def default_filter(record):
    return True


def default_patch(record):
    return record


def get_logger(terminal=True, level=LoggingConfig.DEFAULT_LEVEL, time_format=LoggingConfig.TIME,
               icon_format=LoggingConfig.ICON,
               level_format=LoggingConfig.LEVEL, file_format=LoggingConfig.FILE, function_format=LoggingConfig.FUNCTION,
               message_format=LoggingConfig.MESSAGE,
               logfile=False, logfile_dir=None):
    """

    Get a pre-configured loguru logger, if dependency is present, otherwise default to native logger.

    """

    try:
        from loguru import logger as logger_loguru
        logger = logger_loguru
    except ImportError:
        import logging

        logger = logging.getLogger(None)
        logger.setLevel(level)

        return logger



    components = [time_format, icon_format, level_format, file_format, function_format, message_format]
    format = LoggingConfig.SEP.join([component for component in components if component])
    logger.remove()

    if terminal:
        logger.add(sys.stderr, format=format, level=level, filter=default_filter)
        logger = logger.patch(default_patch)

    if logfile:
        logfile_dir = Path(logfile_dir or '.')
        logfile_path = logfile_dir / LoggingConfig.FILENAME
        logger.add(logfile_path, format=format)

    return logger


logger = get_logger()
